package team.frc.morpheus.robot.subsystems;

import com.revrobotics.CANSparkMax;
import com.revrobotics.CANSparkMaxLowLevel;
import com.revrobotics.CANSparkMax.IdleMode;

import edu.wpi.first.wpilibj.*;
import edu.wpi.first.wpilibj.command.Subsystem;
import jaci.pathfinder.Pathfinder;
import jaci.pathfinder.Trajectory;
import jaci.pathfinder.followers.EncoderFollower;
import jaci.pathfinder.modifiers.TankModifier;
import team.frc.morpheus.robot.RobotMap;
import team.frc.morpheus.util.input.Coordinate;
import team.frc.morpheus.util.input.Filters;

import java.util.Optional;
import java.util.Timer;
import java.util.TimerTask;

/**
 *
 */
public class DriveBase extends Subsystem {
    public static final double WIDTH = 26;
    public static final int DEFAULT_CRUISE_VELOCITY = 685;
    public static final Trajectory.Config CONFIG = new Trajectory.Config(Trajectory.FitMethod.HERMITE_CUBIC,
            Trajectory.Config.SAMPLES_HIGH, 0.05, 5, 5, 60);
    public static final int DEFAULT_ACCELERATION = 611;
    private static final double TURN_SENSITIVITY = 1, KP_GYRO = 0.01;
    private double qsAccumulator = 0;
    private CANSparkMax left, leftFollower, right, rightFollower;
    private Encoder leftEncoder, rightEncoder;
    private Optional<ADXRS450_Gyro> gyro = Optional.empty();
    private Timer trajectoryScheduler;
    private Optional<TimerTask> trajectoryRunner = Optional.empty();
    private static Optional<DriveBase> instance = Optional.empty();
    private int ticksPerRev = 480*4; //PracBot = 346
    private double wheelDiameter = .152; //PracBot = .148

    private DriveBase() {
        instance = Optional.of(this);
        left = new CANSparkMax(RobotMap.LEFT_DRIVE, CANSparkMaxLowLevel.MotorType.kBrushless);
        right = new CANSparkMax(RobotMap.RIGHT_DRIVE, CANSparkMaxLowLevel.MotorType.kBrushless);
        leftFollower = new CANSparkMax(RobotMap.LEFT_FOLLOWER, CANSparkMaxLowLevel.MotorType.kBrushless);
        rightFollower = new CANSparkMax(RobotMap.RIGHT_FOLLOWER, CANSparkMaxLowLevel.MotorType.kBrushless);

        leftEncoder = new Encoder(RobotMap.LEFT_CHANNEL_A, RobotMap.LEFT_CHANNEL_B, RobotMap.INTAKE_SIDE_FORWARD);
        rightEncoder = new Encoder(RobotMap.RIGHT_CHANNEL_A, RobotMap.RIGHT_CHANNEL_B, !RobotMap.INTAKE_SIDE_FORWARD);

        //gyro = new ADXRS450_Gyro();
        //gyro.calibrate();
        //gyro.reset();

        trajectoryScheduler = new Timer();

        left.setInverted(!RobotMap.INTAKE_SIDE_FORWARD);
        right.setInverted(RobotMap.INTAKE_SIDE_FORWARD);

//        leftFollower.follow(left);
//        rightFollower.follow(right);

        leftFollower.setInverted(!RobotMap.INTAKE_SIDE_FORWARD);
        rightFollower.setInverted(RobotMap.INTAKE_SIDE_FORWARD);

        left.setSmartCurrentLimit(80); // default is 80
        leftFollower.setSmartCurrentLimit(80);
        right.setSmartCurrentLimit(80);
        rightFollower.setSmartCurrentLimit(80);

        // TODO: apply pid settings to each controller via getPIDController()

        // /* set closed loop gains in slot */
        // left.selectProfileSlot(0, 0);
        // left.config_kF(0, 1.4, 0);
        // left.config_kP(0, 16, 0);
        // left.config_kI(0, 0, 0);
        // left.config_kD(0, 0, 0);
        // /* zero the sensor */
        // left.setSelectedSensorPosition(0, 0, 0);

        // /* set closed loop gains in slot */
        // right.selectProfileSlot(0, 0);
        // right.config_kF(0, 1.65, 0);
        // right.config_kP(0, 16, 0);
        // right.config_kI(0, 0, 0);
        // right.config_kD(0, 0, 0);
        // /* zero the sensor */
        // right.setSelectedSensorPosition(0, 0, 0);

    }

    public static synchronized DriveBase getInstance() {
        return instance.orElseGet(DriveBase::new);
    }

    public void set(Coordinate pt) {
        cancelTrajectory();
        Filters.constrain(pt);
        // left.setParameter(ConfigParameter.kCtrlType, ControlType.kDutyCycle.value);
        // right.setParameter(ConfigParameter.kCtrlType, ControlType.kDutyCycle.value);

        directSet(pt.y() + pt.x(), pt.y() - pt.x());
        // left.set(pt.y() + pt.x());
        // right.set(pt.y() - pt.x());
    }

    public void setCubic(Coordinate pt) {
        cancelTrajectory();
        Filters.constrain(pt);
        directSet(Math.pow(pt.y() + pt.x(), 3), Math.pow(pt.y() - pt.x(), 3));
    }

    // inspired by
    // https://github.com/Team254/FRC-2016-Public/blob/master/src/com/team254/frc2016/CheesyDriveHelper.java
    public void setCheesy(Coordinate pt, boolean quickTurn) {
        // TRAJ cancelTrajectory();
        Filters.constrain(pt);

        double overPower;
        double angularPower;

        if (quickTurn) {
            if (Math.abs(pt.y()) < 0.2) {
                double alpha = 0.1;
                qsAccumulator = (1 - alpha) * qsAccumulator + alpha * pt.x() * 2;
            }
            overPower = 1.0;
            angularPower = pt.x();
        } else {
            overPower = 0.0;
            angularPower = Math.abs(pt.y()) * pt.x() * TURN_SENSITIVITY - qsAccumulator;
            if (qsAccumulator > 1)
                qsAccumulator -= 1;
            else if (qsAccumulator < -1)
                qsAccumulator += 1;
            else
                qsAccumulator = 0;
        }

        double rightOut = pt.y() - angularPower;
        double leftOut = pt.y() + angularPower;

        if (leftOut > 1) {
            rightOut -= overPower * (leftOut - 1);
            leftOut = 1;
        } else if (rightOut > 1) {
            leftOut -= overPower * (rightOut - 1);
            rightOut = 1;
        } else if (leftOut < -1) {
            rightOut += overPower * (-1 - leftOut);
            leftOut = -1;
        } else if (rightOut < -1) {
            leftOut += overPower * (-1 - rightOut);
            rightOut = -1;
        }

        directSet(leftOut, rightOut);
    }

    public void directSet(double l, double r) {
        if (RobotMap.INTAKE_SIDE_FORWARD) {
            left.set(r);
            leftFollower.set(r); // disable if following is re-enabled
            right.set(l);
            rightFollower.set(l); // disable if following is re-enabled
        } else {
            left.set(l);
            leftFollower.set(l); // disable if following is re-enabled
            right.set(r);
            rightFollower.set(r); // disable if following is re-enabled
        }
    }

    public int getLeftPosition() {
        // return left.getEncoder().getWristPosition();
        if (!RobotMap.INTAKE_SIDE_FORWARD)
            return leftEncoder.get();
        else
            return rightEncoder.get();
    }

    public int getRightPosition() {
        // return right.getEncoder().getWristPosition();
        if (!RobotMap.INTAKE_SIDE_FORWARD)
            return rightEncoder.get();
        else
            return leftEncoder.get();
    }


    public void enableBrakes() {
        left.setIdleMode(IdleMode.kBrake);
        right.setIdleMode(IdleMode.kBrake);
        leftFollower.setIdleMode(IdleMode.kBrake);
        rightFollower.setIdleMode(IdleMode.kBrake);
    }

    public void disableBrakes() {
        left.setIdleMode(IdleMode.kCoast);
        right.setIdleMode(IdleMode.kCoast);
        leftFollower.setIdleMode(IdleMode.kCoast);
        rightFollower.setIdleMode(IdleMode.kCoast);
    }

    public boolean isMoving() {
        return Math.abs(left.getEncoder().getVelocity()) + Math.abs(right.getEncoder().getVelocity()) != 0;
    }

    public void executeTrajectory(Trajectory trajectory, boolean reverse) {
        executeTrajectory(trajectory, reverse, false);
    }

    public void executeTrajectory(Trajectory trajectory, boolean reverse, boolean useGyro) {
        cancelTrajectory();

        TankModifier modifier = new TankModifier(trajectory).modify(WIDTH * 0.0254);
        EncoderFollower leftEF = new EncoderFollower(modifier.getLeftTrajectory());
        EncoderFollower rightEF = new EncoderFollower(modifier.getRightTrajectory());

        if (reverse) {
            /* TODO: huge red flag, getWristPosition returns double! */
            leftEF.configureEncoder(-getRightPosition(), ticksPerRev, wheelDiameter);
            rightEF.configureEncoder(-getLeftPosition(), ticksPerRev, wheelDiameter);
            leftEF.configurePIDVA(1.0, 0, 0, 0.075, 0); // kv was .25
            rightEF.configurePIDVA(1.0, 0, 0, 0.075, 0);//kv was .25
        } else {
            leftEF.configureEncoder(getLeftPosition(), ticksPerRev, wheelDiameter);
            rightEF.configureEncoder(getRightPosition(), ticksPerRev, wheelDiameter);
            leftEF.configurePIDVA(1, 0, 0, 0.075, 0);
            rightEF.configurePIDVA(1, 0, 0, 0.075, 0);
        }

        if (useGyro && gyro.isPresent()) {
            gyro.get().reset();
        }

        trajectoryRunner = Optional.of(new TimerTask() {
            @Override
            public void run() {
                if (!leftEF.isFinished() || !rightEF.isFinished()) {
                    if (reverse) {
                        double l = leftEF.calculate(-getRightPosition());
                        double r = rightEF.calculate(-getLeftPosition());
                        if (useGyro && gyro.isPresent()) {
                            double gyroCorrection = KP_GYRO
                                    * Pathfinder.boundHalfDegrees(leftEF.getHeading() - gyro.get().getAngle());
                            l += gyroCorrection; // signs may need to be flipped in reverse case...
                            r -= gyroCorrection;
                        }
                        if (l > 1 || r > 1) {
                            double d = Math.max(l, r);
                            l = l / d;
                            r = r / d;
                        }
                        directSet(-r, -l);
                    } else {
                        double l = leftEF.calculate(getLeftPosition());
                        double r = rightEF.calculate(getRightPosition());
                        if (useGyro && gyro.isPresent()) {
                            double gyroCorrection = KP_GYRO
                                    * Pathfinder.boundHalfDegrees(leftEF.getHeading() - gyro.get().getAngle());
                            l += gyroCorrection;
                            r -= gyroCorrection;
                        }
                        if (l > 1 || r > 1) {
                            double d = Math.max(l, r);
                            l = l / d;
                            r = r / d;
                        }
                        directSet(l, r);
                    }
                } else {
                    cancelTrajectory();
                }
            }
        });
        trajectoryScheduler.scheduleAtFixedRate(trajectoryRunner.get(), 0, 50);
    }

    public synchronized void cancelTrajectory() {
        if (trajectoryRunner.isPresent()) {
            trajectoryRunner.get().cancel();
            trajectoryRunner = Optional.empty();
            directSet(0, 0); // better to be safe than sorry
        }
    }

    public boolean isTrajectoryInProgress() {
        return trajectoryRunner.isPresent();
    }

    public GyroBase getGyro() {
        if (gyro.isPresent()) return gyro.get();
        else return null; // fix later ;)
    }

    public boolean areBrakesEnabled() {
        return left.getIdleMode() == IdleMode.kBrake && right.getIdleMode() == IdleMode.kBrake &&
                leftFollower.getIdleMode() == IdleMode.kBrake && rightFollower.getIdleMode() == IdleMode.kBrake;
    }

    public void initDefaultCommand() {
        // Set the default command for a subsystem here.
        // setDefaultCommand(new MySpecialCommand());
    }
}
